"""

The QAM Server executes remote method calls received from a QAMProxy.

**The QAMServer consists of the following classes:**

``qam.qam_server.QAMServer`` :
  The *QAMServer* class provides the basic functions for registering methods and class-instances which could be executed and called by the server.
  Further the QAMServer receives messages (remote-method-calls) from an amqp-queue and passes them the ExecuteMethodThread which executes one method-call.
  
``qam.qam_server.ExecuteMethodThread`` :
  The ExecuteMethodThread is called by the QAMServer and executes one method-call. 
  The QAMServer receives the method-calls and then the QAMServer starts an ExecuteMethodThread for executing the method.
  
``qam.qam_server.PublisherThread`` :
  The PublisherThread sends the result for one method-call back to QAMProxy. The resuls are stored in a queue. 
  The PublisherThread takes continuously a result from the queue and sends it. 
  If the queue is empty, the PublisherThread waits, until a new result is available for sending.

**The QAMServer consists of the following Exceptions:**

``qam.qam_server.QAMMEthodNotFoundException`` :
  This exception signals, that the method the QAMProxy wanted to call isn't registered on the QAMServer. The exception is sent back to QAMProxy as result.

=======
Usage
=======
Register methods
----------------
An example-usage of the QAMServer. The code shows how to register a function.

   >>> from qam.qam_server import QAMServer
   >>> qam_server = QAMServer(hostname="localhost",
   ...                    port=5672,
   ...                    username='guest',
   ...                    password='guest',
   ...                    vhost='/',
   ...                    server_id='qamserver')
   ...
   >>> def adder_function(x, y):    
   ...    return x + y
   ...
   >>> qam_server.register_function(adder_function, 'add')
   ...
   ... # it is also possible to register the adder_function as follows:
   ... # qam_server.register_function(adder_function)
   ... # the method-name for registering in this case is adder_function.__name__
   ...
   >>> qam_server.serve()
   
Register class-instances
------------------------
An example-usage of the QAMServer. The code shows how to register class-instance.

   >>> from qam.qam_server import QAMServer
   >>> qam_server = QAMServer(hostname="localhost",
   ...                    port=5672,
   ...                    username='guest',
   ...                    password='guest',
   ...                    vhost='/',
   ...                    server_id='qamserver')
   ...
   >>> class TestClass():
   ...    def __init__(self):
   ...        pass
   ...    def adder_function(self,a,b):
   ...        return a+b
   ...
   >>> testclass = TestClass()
   >>> qam_server.register_class(testclass,'testclass')
   >>> qam_server.serve()


=======
Classes
=======
"""
__author__ = "christian"
__date__ = "$Aug 14, 2009 8:24:39 PM$"

import thread, sys
from threading import Thread
import time

from amqp_credentials import AMQPCredentials
from carrot.connection import AMQPConnection
from carrot.messaging import Consumer
from carrot.messaging import Publisher
import logging

logger = logging.getLogger('qam.qam_server')


class QAMServer:
  """
  The *QAMServer* class provides the basic functions for registering methods and class-instances which could be executed and called by the server.
  Further the QAMServer receives messages (remote-method-calls) from an amqp-queue and passes them to a thread, which executes the remote-method-calls.

  :param hostname: hostname of the amqp-server
  :param port: port of the amqp-server
  :param username: username for amqp-server
  :param password: password for amqp-server
  :param vhost: virtual host on amqp-server
  :param server_id: see ``server_id``


  *Class-variables of QAMServer:*
  
  ``server_id : string``
    id for QAMServer. 
  ``method_dict : dictionary``
    stores method-registrations {method_name: method}
  ``instance_dict : dictionary``
    stores class-instance-registrations {instance_name: instance}
  ``result_queue : list``
    stores result-messages processed from ExecutionThread
  ``kill : bool``
    tells the serve-function to terminate
  ``finished : bool``
    tells the close-function wether the serve-function has finished or not
  ``closed : bool``
    tells the destructor wether the amqp-connection is closed or not
  """
  def __init__(self,
               hostname="localhost",
               port=5672,
               username='guest',
               password='guest',
               vhost='/',
               server_id='testqamserver',
               serializer='pickle'):

    """
    Initially the amqp-connection, a consumer and a publisher-thread are being created.
    The publisher-thread sends messages from a given queue (result_queue) to a specified queue.
    """
    logger.debug('Init QAMServer')
    self.server_id = server_id
    self.method_dict = {}
    self.instance_dict = {}
    self.result_queue = []
    self.kill = False
    self.finished = False
    self.closed = False
    self.serializer = serializer

    self.amqp_credentials = AMQPCredentials(hostname=hostname,
                                            port=port,
                                            username=username,
                                            password=password,
                                            vhost=vhost)

    self.amqpconn = AMQPConnection(hostname=hostname,
                                   port=port,
                                   userid=username,
                                   password=password,
                                   host=vhost)

    self.consumer = Consumer(connection=self.amqpconn,
                             queue=server_id + '_server_queue',
                             exchange=server_id + '_server_exchange',
                             routing_key=server_id + '_server_routing_key')

    self.publisher_thread = PublisherThread(amqp_credentials=self.amqp_credentials,
                                            result_queue = self.result_queue,
                                            server_id= self.server_id,
                                            serializer = self.serializer)
    self.publisher_thread.setDaemon(True)
    self.publisher_thread.start()

  def __del__(self):
    if not self.closed: self.close()

  def register_function(self, method, name=None):
    """
    Methods that should be available on the server for calling, are registered with this function.
    Each method is being stored in a dictionary (method_dict), where the name of the method is the key, and the value is the method itself.

    :param method: method to register on server
    :param name: if a name is given, the method is being registered with this name, else the method is registered with method.__name__
    :type name: String or None
    """

    logger.debug('should register method %s with name %s ' % (method.__name__, name))
    if name != None:
      logger.debug('register method with name %s' % name)
      self.method_dict[name] = method
    else:
      logger.debug('register method with name %s' % method.__name__)
      self.method_dict[method.__name__] = method

  def register_class(self,instance,name):
    """
    Class-instances which should be available for calling, are registered with this function.
    Each class-instance is being stored in a dictionary (instance_dict), where the key is the name given and the value is the instance itself.
    No methods for this instance are stored because they could be reached with the function getattr for any given class.

    :param instance: class-instance which should be registered
    :param name: with this name the instance is being registered
    """
    self.instance_dict[name] = instance

  def close(self):
    """
    The consumer, the publisher and the amqp-connection are being closed.
    The serve-function terminates (kill-signal is set to True).
    """
    if self.closed: return
    
    self.kill = True
    while not self.finished: pass
    self.consumer.close()
    self.publisher_thread.close()
    self.amqpconn.close()
    self.closed = True;
    


  def serve(self):
    """
    The serve-function fetches messages (method-calls) from the amqp-queue.
    Each message is being executed by the ExecuteMethodThread, who calls the function and returns the result.
    """
    while not self.kill:
    #def execute_method(message_data, message):
      message = self.consumer.fetch()
      if message != None:
        message_data = message.payload
        message.ack()
        execute_method_thread = ExecuteMethodThread(self.result_queue,
                                                  message_data,
                                                  self.method_dict,
                                                  self.instance_dict,
                                                  self.serializer)
        execute_method_thread.start()

    self.finished = True



class ExecuteMethodThread(Thread):
  """
  The role of the ExecuteMethodThread is calling defined methods and creating messages out of the results, which could be send.
  Messages are stored in the a queue (result_queue) for further processing.
  
  :param result_queue: see ``result_queue``
  :param message_data: see ``message_data``
  :param method_dict: see ``method_dict``
  :param instance_dict: see ``instance_dict``
  
  *Class-variables of ExecuteMethodThread:*

  ``result_queue : list``
    result_queue from parent QAMServer. Results are stored in this list, after executing a specified method.
  ``message_data : dictionary``
    sturcture of the message.
  ``method_dict : dictionary``
    method_dict from parent QAMServer. All registered methods are stored in here.
  ``instance_dict : dictionary``
    instance_dict from parent QAMServer. All registered instances are stored in here.



  """

  def __init__ (self, result_queue, message_data, method_dict, instance_dict, serializer):
    Thread.__init__(self)
    self.result_queue = result_queue
    self.message_data = message_data
    self.method_dict = method_dict
    self.instance_dict = instance_dict
    self.serializer = serializer
   
  def run(self):
    """
    The run-method takes messages, and processes them.
    Content of a message is a simple method call, or a method-call of a given instance.
    For calling the method, it is necessary, that the message is registered on the QAMServer. A method-call of a non-registered method or instance.method throws an exception.
    The result of the call is packed into a message and stored in a queue (result_queue).

    """
    logger.debug('Start Run')
    method_name = self.message_data['method_name']
    params = self.message_data['params']
    routing_key = self.message_data['routing_key']
    id = self.message_data['id']
    result_record = {}

    try:

      if method_name.find('.') == -1:
        # in case a simple method is being called
        logger.debug('in Method Call')
        if method_name not in self.method_dict:
            logger.debug('MethodNotFoundException should be raised')
            raise MethodNotFoundException()
        result = self.method_dict[method_name](*params)
      else:
        # in case instance.method_name is being called
        execution_list = method_name.split('.')
        instance_name = execution_list[0]
        method_name = execution_list[1]

        if instance_name not in self.instance_dict:
          raise MethodNotFoundException()

        instance = self.instance_dict[instance_name]

        if not hasattr(instance,method_name):
          raise MethodNotFoundException

        result = getattr(instance,method_name)(*params)

    # internal exception
    except MethodNotFoundException, e:
      logger.debug('catched MethodNotFoundException')
      result_message = {'id':id, 'result':'Methodname %s is not registered on QAMServer' % method_name, 'exception':True, 'internal_exception':1}
    # external exception
    except:
      logger.debug('Exception happend')
      t,v =  sys.exc_info()[:2]
      
      sys.exc_clear()

      if self.serializer == 'pickle':
          logger.debug('pickle: catched exception: %s' % repr(v))
          result_message = {'id':id, 'result':v, 'exception':True, 'internal_exception':-1}
      else:
          logger.debug('json: catched exception: %s' % repr(v.args))
          result_message = {'id':id, 'result':v.args, 'exception':True, 'internal_exception':-1}

    # no exception
    else:
      result_message = {'id':id, 'result':result, 'exception':False, 'internal_exception':-1}

    result_record = {'result_message':result_message, 'routing_key':routing_key}
    self.result_queue.append(result_record)
    logger.debug('Method Queue: %s' % repr(self.result_queue))

class PublisherThread(Thread):
  """
  The PublisherThread sends given messages from a queue back to the caller proxy.

  :param amqp_credentials: see ``amqp_credentials``
  :param result_queue: see ``result_queue``
  :param server_id: see ``server_id``

  *Class-variables of PublisherThread:*

  ``amqp_credentials : qam.amqp_credentials.AMQPCredentials``
    all necessary stuff for creating a amqp-connection ist stored in here.
  ``result_queue : list``
    result_queue from parent QAMServer. Messages are taken from this queue and sent back to the caller proxy.
  ``server_id : string``
    id for QAMServer.
  ``amqpconn : carrot.connection.AMQPConnection``
    AMQP-Connection for the PublisherThread
  ``kill : bool``
    The kill signal tells the PublisherThread to stop sending messages.
  """

  def __init__ (self, amqp_credentials, result_queue, server_id, serializer):
    Thread.__init__(self)
    self.amqp_credentials = amqp_credentials
    self.result_queue = result_queue
    self.server_id = server_id
    self.amqpconn = AMQPConnection(hostname=self.amqp_credentials.hostname,
                                   port=self.amqp_credentials.port,
                                   userid=self.amqp_credentials.username,
                                   password=self.amqp_credentials.password,
                                   host=self.amqp_credentials.vhost)
    self.kill = False
    self.serializer = serializer

  def close(self):
    """
    The PublisherThread is being stopped (kill-signal is set to True)
    """
    self.kill = True
    self.join()

  def run(self):
    """
    Messages from the given queue (result_queue) are sent back to a specified amqp-queue.
    """
    while not self.kill:
      if len(self.result_queue) == 0:
        time.sleep(0.001)
        continue
      result_record = self.result_queue.pop(0)
      logger.debug('popped message from queue: %s' % repr(result_record['result_message']))
      routing_key = result_record['routing_key']
      logger.debug('create Publisher')
      logger.debug('Serialization: %s' % self.serializer)
      publisher = Publisher(connection=self.amqpconn,
                            exchange=self.server_id + "_client_exchange",
                            routing_key=routing_key,
                            serializer=self.serializer)
      logger.debug('created Publisher')
      publisher.send(result_record['result_message'])
      logger.debug('sent message: %s' % repr(result_record['result_message']))
      publisher.close()
      logger.debug('publisher closed')


class MethodNotFoundException(Exception):
  """
  Internal QAMServer Exception, which is thrown, if a non-registered method is being called.
  """
  pass





"""QAM is a simple RPC Framework which uses the AMQP Protocoll as Transport Protocoll."""
VERSION = (0, 2, 0)
__version__ = ".".join(map(str, VERSION))
__author__ = "QAM Team"
__contact__ = "qamteam@gmail.com"
__homepage__ = "http://packages.python.org/qam"
__docformat__ = "restructuredtext"




import time
from threading import Thread
from qam.qam_server import QAMServer
from custom_exception import CustomException
import os
import sys

sys.path.insert(0, os.pardir+'/src')
sys.path.append(os.getcwd())


class QAMTestServer(Thread):
  def __init__ (self, server_id, serializer='pickle'):
    Thread.__init__(self)
     # Create server
    self.server = QAMServer(hostname="localhost",
                       port=5672,
                       username='guest',
                       password='guest',
                       vhost='/',
                       server_id=server_id,
                       serializer=serializer)

  def close(self):
    self.server.close()

  def run(self):

   
                       


    # Register a function under a different name
    def adder_function(x, y):
      #print 'I am the adder func'
      return x + y

    def hello(str):
      #time.sleep(3)
      return 'Hello ' + str

    def excp_method(arg, arg2):
      raise Exception(arg, arg2)

    def object_method(arg):
      return arg

    def callback_test_method(str):
      #time.sleep(1)
      return 'Callback test result: ' + str

    def callback_exception(msg):
      raise Exception(msg)

    def looong_lasting_method():
      time.sleep(2)
      return

    def complex_type_method(testcomplex):
        return testcomplex.var

    def complex_method_exception(exception):
        return exception[0]

    def complex_method_custom_exception():
        raise CustomException

    def delayed_exception_method():
        time.sleep(2)
        raise CustomException
    
    self.server.register_function(hello, 'hello')
    self.server.register_function(adder_function, 'add')
    self.server.register_function(excp_method)
    self.server.register_function(object_method)
    self.server.register_function(callback_test_method)
    self.server.register_function(callback_exception)
    self.server.register_function(looong_lasting_method)
    self.server.register_function(complex_type_method)
    self.server.register_function(complex_method_exception)
    self.server.register_function(complex_method_custom_exception)
    self.server.register_function(delayed_exception_method)

    testclass = TestClass()
    self.server.register_class(testclass,'testclass')

    # Run the server's main loop
    self.server.serve()

class TestClass():
  def __init__(self):
    pass

  def test_method(self,a,b):
    return a+b



"""
The QAMProxy sends method-calls to a QAMServer-instance and receives the results.
It is possible to register a callback-method on the proxy.
This method is  called when a result for a specific method-call is available.

**The QAMProxy consists of the following classes:**

``qam.qam_proxy.QAMProxy`` :
  This class manages the start of:
  
    - ConsumerThread
    - PublisherThread
    - CallbackObserverThread

  The PublisherThread runs in an infinite-loop and sends continuously messages to the QAMServer.
  The messages which are send to the QAMServer are stored in a method_queue.

  The ConsumerThread receives continuously messages from the QAMServer and signals that a message has arrived.
  If a callback is registered the ConsumerThread signals the CallbackObserverThread that a message has arrived, else it signals the QAMProxy that the message has arrived.

  The CallbackObserverThread checks continuously if it is time to execute a callback-method.
  A callback-method is executed if a callback-method is registered and the ConsumerThread signals that a result has arrived.
``qam.qam_proxy.ConsumerThread`` :
  The ConsumerThread receives continuously messages from the QAMServer and signals that a message has arrived.
  Therefor it is necessary to keep track of all remote-method-calls ever send to the QAMServer.
  The ConsumerThread keeps a list with all QAMServer calls, that haven't received a result yet, so it is possible to relate a result to a caller.
  This list consists of "slots". A slot is identified by an id.
  The slot is reserved if the caller is still waiting on a message and the slot with the given id is being freed if the result has arrived.
``qam.qam_proxy.PublisherThread`` :
  The PublisherThread runs in an infinite-loop and sends continuously messages to the QAMServer.
  Therefor he keeps a queue (``method_queue``) so that every method which is not sent yet, isn't loosed.
  If the queue is empty, the PublisherThread waits until a new message arrives the queue.
``qam.qam_proxy.CallbackObserverThread`` :
  The CallbackObserverThread checks continuously if it is time to execute a callback-method.
  If a callback-method is registered and a result arrived in the ConsumerThread, the CallbackObserverThread will be informed about it.
  Then the he checks wether an error occourred on the QAMServer or not.
  If no error occoured the callback-method which is registered for this request is being called with the result from the QAMServer as parameter.
  If an error occoured and a error-method is registered this method will be called, with the exception as parameter.
  Else a log-message will be written into the logfiles.
  It is the job of the ExecuteCallbackThread to call the callback-methods, otherwise the CallbackObserverThread would block.

``qam.qam_proxy.ExecuteCallbackThread`` :
  The ExecuteCallbackThread receives a method and parameters from the CallbackObserverThread.
  The job of the ExecuteCallbackThread is to execute this method.

``qam.qam_proxy.Callback`` :
  This Class is used to realize the callback-mechanism.
``qam.qam_proxy._Method`` :
  This Class is used to realize the remote-method-call.


**The QAMProxy consists of the following Exceptions:**

``qam.qam_proxy.QAMException`` :
  This is the class for internal exceptions. An internal exception would be everything raised by the QAMServer.
``qam.qam_proxy.QAMMEthodNotFoundException`` :
  This exception is created, if the method called by the QAMProxy isn't registerd on the QAMServer.

=======
Usage
=======

Calling a method without callback
---------------------------------
An example-usage of the QAMProxy. The code shows how to call a method that is registered on QAMServer.

   >>> from qam.qam_proxy import QAMProxy, QAMMethodNotFoundException, QAMException
   >>> qam_proxy = QAMProxy(hostname="localhost",
   ...                    port=5672,
   ...                    username='guest',
   ...                    password='guest',
   ...                    vhost='/',
   ...                    server_id='qamserver',
   ...                    client_id='qamproxy')
   ...
   >>> result = qam_proxy.add(2,3) # name of method registered on QAMServer
   >>> qam_proxy.close()
   
Calling a method with callback
------------------------------
An example-usage of the QAMProxy with callback functions.

   >>> from qam.qam_proxy import QAMProxy, QAMMethodNotFoundException, QAMException
   >>> qam_proxy = QAMProxy(hostname="localhost",
   ...                    port=5672,
   ...                    username='guest',
   ...                    password='guest',
   ...                    vhost='/',
   ...                    server_id='qamserver',
   ...                    client_id='qamproxy')
   ...
   ... # defining functions for callback
   >>> def success(arg):
   ...    print arg
   >>> def error(arg):
   ...    # if an error occours on QAMServer
   ...    print arg 
   >>> uid = qam_proxy.callback(success, error).adder_function(2,4)
   >>> while True:
   ...    state = qam_proxy.get_callback_state(uid)
   ...    if state == 2 :
   ...        # execution of callback finished
   ...        break
   ...
   >>> qam_proxy.close()
   
 
Calling a registered Class
------------------------------
An example-usage of the QAMProxy for calling a registered class-instance on QAMServer.

   >>> from qam.qam_proxy import QAMProxy, QAMMethodNotFoundException, QAMException
   >>> qam_proxy = QAMProxy(hostname="localhost",
   ...                    port=5672,
   ...                    username='guest',
   ...                    password='guest',
   ...                    vhost='/',
   ...                    server_id='qamserver',
   ...                    client_id='qamproxy')
   ...
   >>> result = qam_proxy.testclass.adder_function(2,4)
   >>> qam_proxy.close()

=======
Classes
=======
"""
import copy
import logging
import thread
from threading import Thread
import time
import uuid

from amqp_credentials import AMQPCredentials
from carrot.connection import AMQPConnection
from carrot.messaging import Consumer
from carrot.messaging import Publisher

logger = logging.getLogger('qam.qam_proxy')


#===========================================================================
class QAMProxy:
    """
    The role of the QAMProxy is sending messages to a specific QAMServer.
    Instead of waiting on the message it is also possible to register a callback-method which is being called if a result from QAMServer has arrived.
    The QAMProxy manages the start of ConsumerThread, PublisherThread an CallbackObserverThread.

    :param server_id: id of the QAMServer, for reaching the right queue
    :param hostname: hostname where amqp-server is running
    :param port: port for amqp-server
    :param username: username for amqp-server
    :param password: password for amqp-server
    :param vhost: virtual host of amqp-server
    :param client_id: see ``client_id``

    *Class-variables of QAMProxy:*

    ``amqp_credentials : qam.amqp_credentials.AMQPCredentials``
        all necessary stuff for creating a amqp-connection ist stored in here.
    ``consumer_thread : qam.qam_proxy.ConsumerThread``
        Thread for receiving messages from QAMServer.
    ``publisher_thread : qam.qam_proxy.PublisherThread``
        Thread for sending messages to a QAMServer
    ``callback_observer_thread : qam.qam_proxy.CallbackObserverThread``
        Thread for calling the callback-functions when a result is available
    """
    closed = False
    def __init__(self,
                 server_id,
                 hostname="localhost",
                 port=5672,
                 username='guest',
                 password='guest',
                 vhost='/',
                 client_id=None,
                 serializer='pickle'):
                 
        if client_id == None:
            self.client_id = str(uuid.uuid4())
        else:
            self.client_id = client_id

        logger.info('Init QAMProxy')
        self.serializer = serializer
    


        self.amqp_credentials = AMQPCredentials(hostname=hostname,
                                                port=port,
                                                username=username,
                                                password=password,
                                                vhost=vhost)

   
        self.consumer_thread = ConsumerThread(amqp_credentials=self.amqp_credentials,
                                              server_id=server_id,
                                              client_id=self.client_id,
                                              serializer=self.serializer)

        self.publisher_thread = PublisherThread(amqp_credentials=self.amqp_credentials,
                                                server_id=server_id,
                                                client_id=self.client_id,
                                                serializer=self.serializer)
   
        self.callback_observer_thread = CallbackObserverThread(self)


        self.consumer_thread.setDaemon(True)
        self.publisher_thread.setDaemon(True)
        self.callback_observer_thread.setDaemon(True)
    
        self.consumer_thread.start()
        self.publisher_thread.start()
        self.callback_observer_thread.start()

    def __del__(self):
        self.close()

    def close(self):
        """
        The PublisherThread, the ConsumerThread and the CallbackObserverTrhead are being stopped (kill-signal set to True).
        """
        if self.closed == True: return
        
        if self.publisher_thread != None:
            self.publisher_thread.exit()
        if self.consumer_thread != None:
            self.consumer_thread.exit()
        if self.callback_observer_thread != None:
            self.callback_observer_thread.exit()
        self.closed = True
    
    def callback(self, callback_method, exception_method=None):
        """
        The callback-function registers a callback-method and an exception-method if an exception-method is given.

        :param callback_method: the function that should be called
        :param exception_method: exception_method
        :type exception_method: function or None
        :type callback_method: function
        :rtype: callable ``qam.qam_proxy.Callback``-instance
        """
        if self.callback_method is not None and self.exception_method is None:
            logger.warning('Method has NO exception_method defined. \
                      In case of an Exception on the Server you WILL NOT\
                      get notified!')
        id = self.consumer_thread.get_free_result_list_id(callback_method, exception_method)
        return Callback(self, id, self.client_id)

    def set_timeout(self, timeout):
        """
        The set_timeout-function sets a timeout in seconds and returns a callable
        class where the remote function can be called. If a timeout occurs a
        QAMTimeoutException is raised.

        :param timeout: timeout value in seconds
        :type timeout: integer
        :rtype: callable ``qam.qam_proxy.SynchronTimeout``-instance
        """
        
        return SynchronTimeout(self, self.client_id, timeout)


    def get_num_waiting_callbacks(self):
        """
        This function returns the number of callacks which are waiting for a result from QAMServer.

        :rtype: number of callbacks not executed yet
        """

        return len(self.callback_observer_thread.callback_uid_dict)

    def get_callback_state(self, uid):
        """
        Returns the state of the callback method registered for a specific uid.

        :rtype: state of the callback:

          - 0 = waiting for processing
          - 1 = processing
          - 2 = finished
        """
        item = self.callback_observer_thread.get_callback_item(uid)
        if item is None:
            state = 2
        else:
            state = item['state']
        return state
  

    def __request(self, methodname, params):
        """
        The remote-method-call is packed into a message and the message is stored in a sending-queue.
        A PublisherThread sends the messages to the AMQPServer.
        This function waits, until a result from the QAMServer arrives.

        :param methodname: name of the method that should be executed on the QAMServer
        :param params: parameter for the remote-method-call
        :type methodname: string
        :type param: list of parameters
        :rtype: result received from QAMServer
        """
        id = self.consumer_thread.get_free_result_list_id()
        #assemble message structure (dict) and send message
        self.publisher_thread.add_message_to_send({'id':id,
                                                  'method_name': methodname,
                                                  'routing_key': self.client_id + '_client_routing_key',
                                                  'params': params})

        #loop until result is in list and then return
        while True:
            if self.consumer_thread.is_result_received(id):
                return self.consumer_thread.get_result(id)
            else:
                time.sleep(0.001)

    def __getattr__(self, name):
        """
        This method is invoked, if a method is being called, which doesn't exist on QAMProxy.
        It is used for RPC, to get the function which should be called on the QAMServer.
        """
        # magic method dispatcher
        return _Method(self.__request, name)


class Callback:
    """
    The class Callback is callable. It is used to realize the callback-mechanism.

    :param parent: instance of QAMProxy
    :param id: id of slot in the result-list
    :param client_id: used for routing-key generation
    """
    def __init__(self, parent, id, client_id):
        self.id = id
        self.client_id = client_id
        self.parent = parent
        self.timeout = None

    def set_timeout(self, timeout):
        """
        This method is for setting a timeout for a asynchronous remote method call.
        If the timeout happens before the callback function is called a
        QAMTimeoutException will be passed to the registered error function.

        :param timeout: the timeout in seconds
        :type timeout: integer

        """
        self.timeout = timeout
        return self
        
    
    def __request(self, methodname, params):
        """
        This function is only used, if a callback-function is registered on QAMProxy. The original remote-method-call is packed into a message and stored in a sending-queue.
        A PublisherThread sends the message to the AMQP-Server.
        The CallbackObserverThread is being told to wait for a result from the QAMServer and calls the callback-function when a result arrives.

        :param methodname: name of the method that should be called on the QAMServer
        :param params: parameters for the remote-method-call
        :type methodname: string
        :type params: list of parameters

        :rtype: an uid to identify the acutal callback, necessary for telling the actual state of the callback (waiting,processing,finished)
        """
   

        #assemble message structure (dict) and send message
        self.parent.publisher_thread.add_message_to_send({'id':self.id,
                                                         'method_name': methodname,
                                                         'routing_key': self.client_id + '_client_routing_key',
                                                         'params': params})


        uid = self.parent.callback_observer_thread.init_waiting_callback_item(self.id, self.timeout)
    
        return uid
    
    def __getattr__(self, name):
        # magic method dispatcher
        """
        This method is invoked, if a method is being called, which doesn't exist in the Callback-class.
        It is used for RPC, to get the function which should be called on the QAMServer.
        """
        return _Method(self.__request, name)

#===========================================================================
class SynchronTimeout:
    """
    The class SynchronTimeout is callable. It is used to realize the synchronous
    Timeout-mechanism.

    :param parent: instance of QAMProxy
    :param client_id: used for routing-key generation
    :param timeout: timeout in seconds
    """
    def __init__(self, parent, client_id, timeout):
        self.client_id = client_id
        self.parent = parent
        self.timeout = timeout

    def __request(self, methodname, params):
        """
        The remote-method-call is packed into a message and the message is stored in a sending-queue.
        A PublisherThread sends the messages to the AMQPServer.
        This function waits, until a result from the QAMServer arrives or
        if a timeout occours a QAMTimeoutException gets raised.

        :param methodname: name of the method that should be executed on the QAMServer
        :param params: parameter for the remote-method-call
        :type methodname: string
        :type param: list of parameters
        :rtype: result received from QAMServer
        """
        id = self.parent.consumer_thread.get_free_result_list_id()
        #assemble message structure (dict) and send message
        self.parent.publisher_thread.add_message_to_send({'id':id,
                                                         'method_name': methodname,
                                                         'routing_key': self.client_id + '_client_routing_key',
                                                         'params': params})
        start_time = time.time()
        #loop until result is in list and then return, or if timeout occours
        #raise QAMTimeoutException
        while True:
            curr_time = time.time()
            if curr_time - start_time > self.timeout:
                self.parent.consumer_thread.free_result_item(id)
                raise QAMTimeoutException('Timeout happened during a call of %s' % methodname)
            if self.parent.consumer_thread.is_result_received(id):
                return self.parent.consumer_thread.get_result(id)
            else:
                time.sleep(0.001)

    def __getattr__(self, name):
        # magic method dispatcher
        """
        This method is invoked, if a method is being called, which doesn't exist in the SynchronTimeout-class.
        It is used for RPC, to get the function which should be called on the QAMServer.
        """
        return _Method(self.__request, name)

#===========================================================================

class _Method:
    """
    The _Method-class is used to realize remote-method-calls.
    :param send: name of the function that should be executed on QAMProxy
    :param name: name of the method which should be called on the QAMServer
    """
    # some magic to bind an XML-RPC method to an RPC server.
    # supports "nested" methods (e.g. examples.getStateName)
    def __init__(self, send, name):
        self.__send = send
        self.__name = name
    def __getattr__(self, name):
        return _Method(self.__send, "%s.%s" % (self.__name, name))
    def __call__(self, * args):
        return self.__send(self.__name, args)

#===========================================================================
class ConsumerThread(Thread):
    """
    The ConsumerThread receives messages from the QAMServer.
    When a message arrives, the result is stored in the result_list, identified by the id.
    Further the specific entry in the result_list gets the state *DONE*.

    Structure of the received message:
    
        - "id": integer
        - "result": result of method-call
        - "exception": bool
        - "internal_exception": integer


    :param amqp_credentials: all necessary stuff for creating a amqp-connection ist stored in here.
    :param server_id: see ``server_id``
    :param client_id: see ``client_id``

    *Class-variables of ConsumerThread:*

    ``free_index_queue : list``
        Free slots (ids) from the result_list are stored in the free_index_queue.
    ``result_list : list``
        The result_list stores the result and a state for a remote-method-call identified by an id.
        Before receiving the result the specific slot for a remote-method-call in the result_list has the state *WAITING*.
        After receiving the slot gets the state *DONE*.
    ``server_id : string``
        server_id received from parent QAMProxy. Used for amqp-connection to the server.
    ``client_id : string``
        client_id received from parent QAMProxy. Used for defining routing-key and amqp-queue-name
    ``kill : bool``
        The kill signal tells the ConsumerThread to stop receiving messages.
    ``amqpconn : carrot.connection.AMQPConnection``
        AMQPConnection for the ConsumerThread
    ``consumer : carrot.messaging.Consumer``
        AMQPConsumer for the ConsumerThread. The AMQPConsumer receives the messages sent by the QAMServer
    """
    (WAITING, DONE, FREE) = range(3)
    def __init__(self, amqp_credentials, server_id, client_id, serializer):
        Thread.__init__(self)
        self.free_index_queue = []
        self.result_list = []
        self.server_id = server_id
        self.client_id = client_id
        self.kill = False
        self.serializer = serializer

        self.amqpconn = AMQPConnection(hostname=amqp_credentials.hostname,
                                       port=amqp_credentials.port,
                                       userid=amqp_credentials.username,
                                       password=amqp_credentials.password,
                                       host=amqp_credentials.vhost)

        self.consumer = Consumer(connection=self.amqpconn,
                                 queue=client_id + '_client_queue',
                                 exchange=server_id + '_client_exchange',
                                 routing_key=client_id + '_client_routing_key')

    def get_result_item(self, id):
        """
        This function returns the result from the result_list, for a specific id.

        :param id: id for the desired result_list "slot"
        :type id: integer
        :rtype: result for a specific id
        """
        return self.result_list[id]

    def get_result(self, id):
        """
        This function is a high level function for getting the result of
        the remote method call. It makes several checks to validate the result
        against Exceptions and raises them if any occoured

        Exceptions:
            - QAMMethodNotFoundException: gets raised when the method
              on the server is not found
            - Exception: is raised when a user defined exception is raised on
              the server.


        :param id: id for the desired result
        :type id: integer
        :rtype: result for a specific id
        """
        result_item = self.get_result_item(id)
        result = result_item['result_message']
        self.free_result_item(id)

        if result['exception'] == True:
            if result['internal_exception'] > 0:
                raise QAMMethodNotFoundException(result['result'])
            else:
                if self.serializer == 'pickle':
                    logger.debug("Consumerthread: pickle result: %s" % repr(result['result']))
                    raise result['result']
                else:
                    logger.debug("ConsumerThread: json result: %s " % repr(result['result']))
                    raise Exception(*result['result'])
        else:
            return result['result']

    def is_result_received(self, id):
        """
        This function checks, if a desired result from the QAMServer is already received.

        :param id: id for checking the state in the result_list.
        :type id: integer
        :rtype: True if the message is already received, else False
        """
        if self.result_list[id]['state'] == ConsumerThread.DONE:
            return True
        else:
            return False

    
    def free_result_item(self, id):
        """
        This function frees a specific result_list "slot".

        :param id: id of the slot to free
        :type id: integer
        """
        self.result_list[id]['state'] = ConsumerThread.FREE
        self.free_index_queue.append(id)

    def get_free_result_list_id(self, callback_method=None, exception_method=None):
        """
        This method assigns a free slot to a new request. The new request is appended to result_list with the state *WAITING*.

        :param callback_method: name of the method to call when result arrives
        :param exception_method: name of the method to call if an exception occured on the QAMServer
        :type callback_method: function or None
        :type exception_method: function or None
        :rtype: id of the assigned slot
        """

        blank_record = {'result_message':None,
            'state':ConsumerThread.WAITING,
            'callback_method': callback_method,
            'exception_method': exception_method
            }

        queue_lock = thread.allocate_lock()
        queue_lock.acquire()
   
        if len(self.free_index_queue) > 0:
            id = self.free_index_queue.pop(0)
            queue_lock.release()
            self.result_list[id] = blank_record
        else:
            queue_lock.release()
            lock = thread.allocate_lock()
            lock.acquire()
            self.result_list.append(blank_record)
            id = len(self.result_list) - 1
            lock.release()
        return id

    def exit(self):
        """
        The exit method terminates the ConsumerThread and closes the amqp-consumer and the amqp-connection of the ConsumerThread.
        """
        self.kill = True
        self.join()
        self.consumer.close()
        self.amqpconn.close()
    
    def run(self):
        """
        Messages are received from QAMServer and stored in the result_list. T
        he state in the result_list for specific results is set *DONE*.
        """
        while not self.kill:
            message = self.consumer.fetch()
            if message != None:
                message.ack()
                message_data = message.payload
                result_id = message_data['id']

                if self.result_list[result_id]['state'] != ConsumerThread.FREE:
                    self.result_list[result_id]['result_message'] = message_data
                    self.result_list[result_id]['state'] = ConsumerThread.DONE
#      else:
#        time.sleep(0.001)

#===========================================================================
class PublisherThread(Thread):
    """
    The PublisherThread takes messages from the method_queue and sends them to the QAMServer.
    The method_queue is filled by ``qam.qam_proxy.QAMProxy.__request`` and by ``qam.qam_proxy.Callback.__request``.

    :param amqp_credentials: see ``amqp_credentials``
    :param server_id: see ``server_id``
    :param client_id: see ``client_id``

    *Class-variables of PublisherThread:*

    ``amqp_credentials : qam.amqp_credentials.AMQPCredentials``
        all necessary stuff for creating a amqp-connection ist stored in here.
    ``server_id : string``
        server_id received from parent QAMProxy. Used for amqp-connection to the server.
    ``client_id : string``
        client_id received from parent QAMProxy. Used for defining routing-key and amqp-queue-name
    ``kill : bool``
        The kill signal tells the PublisherThread to stop receiving messages.
    ``method_queue : list``
        All requests for a method-call on the QAMServer are stored in the method_queue.
    ``publisher : carrot.messaging.Publisher``
        The publisher sends messages to the QAMServer.
    """

    def __init__(self, amqp_credentials, server_id, client_id, serializer):
        Thread.__init__(self)
        self.server_id = server_id
        self.client_id = client_id
        self.method_queue = []
        self.amqp_credentials = amqp_credentials
        self.kill = False
        self.serializer = serializer

    def add_message_to_send(self, message):
        """
        A new message is appended to the method_queue. All messages that are stored in the method_queue are sent to the QAMServer.

        :param message: the message which should be appended.
        :type message: dictionary

        The structure of the message is as follows:
          - "id": integer
          - "method_name": string
          - "routing_key": string
          - "params": list
        """
        self.method_queue.append(message)
  
    def exit(self):
        """
        The exit method terminates the PublisherThread and closes the publisher and the amqp-connection
        """
        self.kill = True
        self.join()
        self.publisher.close()
        self.amqpconn.close()

    def run(self):
        """
        The run method sends messages which are stored in the method_queue to the QAMServer.
        """
        self.amqpconn = AMQPConnection(hostname=self.amqp_credentials.hostname,
                                       port=self.amqp_credentials.port,
                                       userid=self.amqp_credentials.username,
                                       password=self.amqp_credentials.password,
                                       host=self.amqp_credentials.vhost)

        logger.debug('Serialization: %s' % self.serializer)
        self.publisher = Publisher(connection=self.amqpconn,
                                   exchange=self.server_id + "_server_exchange",
                                   routing_key=self.server_id + "_server_routing_key",
                                   serializer=self.serializer)
    
        while self.kill == False:
            if len(self.method_queue) == 0:
                time.sleep(0.001)
                continue
            method_message = self.method_queue.pop(0)
            self.publisher.send(method_message)
   
    
#===========================================================================

class CallbackObserverThread(Thread):
    """

    The CallbackObserverThread checks continuosly if for an registered callback-function a result arrived.
    If the result arrived the specific callback-function is being executed by the ExecuteCallbackThread.

    :param parent: see ``parent``

    *Class-variables of CallbackObserverThread:*

    ``parent : qam.qam_proxy.QAMProxy``
        the QAMProxy is necessary to reach the methods of ``qam.qam_proxy.ConsumerThread``
    ``callback_uid_dict : dictionary``
        Dictionary which monitors the callbacks, that are not executed yet.
          Structure of the callback_uid_dict:
            - "id": string (id in the result_list)
            - "state": integer
    ``callback_dickt_lock : threading.Lock``
        A lock for locking the critical sections.
    ``kill : bool``
        The kill signal tells the CallbackObserverThread to stop calling callback-methods.
    """

    def __init__(self, parent):
        Thread.__init__(self)
        self.callback_uid_dict = {}
        self.callback_dict_lock = thread.allocate_lock()
        self.kill = False
        self.parent = parent
    
    def get_callback_item(self, uid):
        """
        This function returns an item from the callback_uid_dict, identified by the given uid.

        :param uid: uid of the desired item
        :type uid: string
        :rtype: dictionary, {"id": string, "state": integer}

        It is a critical section so it is necessary to lock this section.
        """
        
        self.callback_dict_lock.acquire()
        if uid in self.callback_uid_dict:
            item = self.callback_uid_dict.get(uid)
            cpy_item = copy.deepcopy(item)
            #logger.debug('copy' + repr(item))
        else:
            #logger.debug('copy none')
            cpy_item = None
        self.callback_dict_lock.release()
        return cpy_item

    def change_callback_item_state(self, uid, state):
        """
        This method changes the state of a specific callback-item.
        Possible values for the state are:
          - 0 = waiting for processing
          - 1 = processing
          - 2 = finished

        :param uid: uid of the desired item
        :param state: state to be set
        :type uid: string
        :type state: integer

        It is a critical section so it is necessary to lock this section.
        """
        self.callback_dict_lock.acquire()
        item = self.callback_uid_dict[uid]
        item['state'] = state
        self.callback_dict_lock.release()

    def remove_callback_item(self, uid):
        """
        This method removes an item from the callback_uid_dict. This is the case, when a callback is already executed.

        :param uid: uid of the desired item
        :type uid: string

        It is a critical section so it is necessary to lock this section.
        """
        self.callback_dict_lock.acquire()
        del self.callback_uid_dict[uid]
        self.callback_dict_lock.release()

    def __generate_uid(self):
        """
        A new unique uid for callback-registration is generated.

        :rtype: string which contains the generated uid
        """
        while True:
            logger.debug('Search for uid')
            uid = str(uuid.uuid4())
            if uid not in self.callback_uid_dict: break
        return uid

    def init_waiting_callback_item(self, id, timeout):
        """
        This method initializes a new callback-item with an unique uid. The given id is being stored and the state is set to *WAITING*.
        The new item is stored in the callback_uid_dict.

        :param id: id of the item in the result_list
        :type id: integer
        """
        uid = self.__generate_uid()
        self.callback_uid_dict[uid] = {'id':id,
            'state':0,
            'timeout':timeout,
            'start_time': time.time()}
        logger.debug('init uid done')
        return uid

    def check_timeout(self, uid, callback_info):
        """
        This method checks if a timeout has occoured and calls the error method
        with a QAMTimeoutException

        :param uid: uid of the callback info in callback_uid_dict
        :type uid: integer
        :param callback_info: the corresponding info from the callback_uid_dict
        :type cakkback_info: dict
        """
        #check timeout
        id = callback_info['id']
        curr_time = time.time()
        
        if callback_info['timeout'] == None: return False

        if curr_time - callback_info['start_time'] > callback_info['timeout']:
            self.change_callback_item_state(uid, 1)
            result_item = self.parent.consumer_thread.get_result_item(id)
            exception_method = result_item['exception_method']
            self.parent.consumer_thread.free_result_item(id)
            if exception_method == None:
                logger.error('Remote Exception not cought due to exception_method = None')
                self.remove_callback_item(uid)
            else:
                #logger.debug('Start Timeout exception, List: %s' % repr(self.callback_uid_dict))
                exception = QAMTimeoutException('Timeout for callback method occoured.')
                exec_timeout_thread = ExecuteCallbackThread(self.parent,
                                                            exception_method,
                                                            exception,
                                                            uid)
                exec_timeout_thread.start()
                
            return True
        else:
            return False

    def exit(self):
        """
        The exit method terminates the CallbackObseverThread.
        """
        self.kill = True
        self.join()

    def run(self):
        """
        The run method checks for every registered callback, if the result has already arrived. If the result has arrived:
          - in case of success the registered callback-function is being called
          - in case of an error:

            - if an error-function is registered this function is being called with the exception as parameter
            - if no error-function is registered a log-message is written into logfile

        For the execution of the callback-function a new ExecuteCallbackThread is being started. This thread finally executes the method.
        """
        while self.kill == False:
            #logger.debug("In Loop")
            for uid, val in self.callback_uid_dict.items():
                if val['state'] != 0: continue
                id = val['id']

                if self.check_timeout(uid, val): continue

                # check if result is already here
                if self.parent.consumer_thread.is_result_received(id):
                    self.change_callback_item_state(uid, 1)
                    result_item = self.parent.consumer_thread.get_result_item(id)
                    result = result_item['result_message']
                    callback_method = result_item['callback_method']
                    exception_method = result_item['exception_method']
                    self.parent.consumer_thread.free_result_item(id)

                    # check if exception_method is defined
                    if result['exception'] == True:
                        logger.debug('Callback Observer: Exception happened')
                        if exception_method == None:
                            logger.error('Remote Exception not cought due to exception_method = None')
                            self.remove_callback_item(uid)
                        else:
                            # check if it is an internal QAM Exception
                            if result['internal_exception'] > 0:
                                exception = QAMMethodNotFoundException(result['result'])
                                exec_error_thread = ExecuteCallbackThread(self.parent,
                                                                          exception_method,
                                                                          exception,
                                                                          uid)
                                exec_error_thread.start()
                            # else it is a user defined exception
                            else:
                                logger.debug("CallbackObserver: user defined exception: %s " % repr(result['result']))
                                exec_error_thread = ExecuteCallbackThread(self.parent,
                                                                          exception_method,
                                                                          result['result'],
                                                                          uid)
                                exec_error_thread.start()
                    # everything is ok no exception happened
                    else:
                        logger.debug("Callback SUCCESS %s" % repr(result['result']))
                        exec_succes_thread = ExecuteCallbackThread(self.parent,
                                                                   callback_method,
                                                                   result['result'],
                                                                   uid)
                        exec_succes_thread.start()
            
            time.sleep(0.001)

#===========================================================================

class ExecuteCallbackThread(Thread):
    """
    The ExecuteCallbackThread executes a registered callback-function with the results received from the QAMServer as parameter.

    :param parent: see ``parent``
    :param method: see ``method``
    :param params: see ``params``
    :param uid: see ``uid``

    *Class-variables of ExecuteCallbackThread:*

    ``parent : qam.qam_proxy.QAMProxy``
        the QAMProxy is necessary to reach the methods of ``qam.qam_proxy.CallbackObserverThread``:
    ``method : function``
        The method that should be called by the ExecuteCallbackThread.
    ``params : string``
        result received from QAMServer.
    ``uid : string``
        The uid which identifies the actual callback.
    """

    def __init__ (self, parent, method, params, uid):
        Thread.__init__(self)
        self.method = method
        self.params = params
        self.uid = uid
        self.parent = parent
   
    def run(self):
        """
        The run method executes the given callback-function.
        """
       
        self.method(self.params)
        self.parent.callback_observer_thread.remove_callback_item(self.uid)

class QAMException(Exception):
    """
    Internal QAMServer Exception.
    """
    pass

class QAMMethodNotFoundException(QAMException):
    """
    Internal QAMServer Exception, which is generated if the method called by the QAMProxy doesn't exist on the QAMServer.
    """
    pass

class QAMTimeoutException(QAMException):
    """
    Internal QAMServer Exception, which is generated if the method called by the QAMProxy produces a timeout.
    """





class AMQPCredentials:
  def __init__(self,
               hostname="localhost",
               port=5672,
               username='guest',
               password='guest',
               vhost='/',):
    self.hostname = hostname
    self.port = port
    self.username = username
    self.password = password
    self.vhost = vhost


# To change this template, choose Tools | Templates
# and open the template in the editor.

import logging
import logging.handlers
import os
import sys
from threading import Thread
import time
import unittest


sys.path.insert(0, os.pardir + '/src')
sys.path.append(os.getcwd())


from qam.qam_proxy import   QAMProxy, \
                            QAMMethodNotFoundException, \
                            QAMException, \
                            QAMTimeoutException
from qamtestserver import QAMTestServer

from custom_exception import CustomException

# setup logging
formatter = logging.Formatter("%(asctime)s|%(name)s|%(levelname)s|%(filename)s|%(funcName)s| #%(lineno)d: %(message)s")

logger_proxy = logging.getLogger('qam.qam_proxy')
handler = logging.handlers.RotatingFileHandler('log/qamproxy.log', maxBytes=200000, backupCount=5)
handler.setFormatter(formatter)
logger_proxy.addHandler(handler)
logger_proxy.setLevel(logging.DEBUG)

logger_server = logging.getLogger('qam.qam_server')
handler = logging.handlers.RotatingFileHandler('log/qamserver.log', maxBytes=200000, backupCount=5)
handler.setFormatter(formatter)
logger_server.addHandler(handler)
logger_server.setLevel(logging.DEBUG)

logger = logging.getLogger('qam.qam_test')
handler = logging.handlers.RotatingFileHandler('log/qamtest.log', maxBytes=200000, backupCount=5)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)
# end setup logging


class ComplexTest:
    def __init__(self):
        self.var = 'hallo'


class  QamTestCase(unittest.TestCase):

 

    def setUp(self):
        print "Init Testing..."
        self.amqp_hostname = "localhost"
        self.amqp_port = 5672
        self.amqp_user = 'guest'
        self.amqp_pwd = 'guest'
        self.amqp_vhost = '/'

        self.server = QAMTestServer(server_id='testqamserver',serializer='json')
        self.server.setDaemon(True)
        self.server.start()

        self.server_pickle = QAMTestServer(server_id='testqamserverpickle')
        self.server_pickle.setDaemon(True)
        self.server_pickle.start()
    

    def tearDown(self):
        self.server.close()
        self.server_pickle.close()
    #    self.foo.dispose()
    #    self.foo = None

    def test_basic_method_call(self):
        print 'test_basic_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testqamclient1',
                         serializer='json'
                         )

        for i in range(10):
            a = i
            b = 1
            golden_device = a + b
            result = proxy.add(a, b)
            self.assertEqual(golden_device, result, "%d + %d must equal %d" % (a, b, golden_device))

        proxy.close()

    def test_string_method_call(self):
        print 'test_string_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testqamclient2',
                         serializer='json'

                         )

        str = 'Teststring'
        golden_device = 'Hello Teststring'
        result = proxy.hello(str)
        self.assertEqual(golden_device, result, "Golden Device %s must equal result %s" % (golden_device, result))
        proxy.close()



    def test_multiple_clients(self):
        print 'test_multiple_clients'
        class TestClientThread(Thread):
            def __init__(self, client_id, test):
                Thread.__init__(self)
                self.client_id = client_id
                self.failed = False
                self.test = test
            def run(self):
                proxy = QAMProxy(hostname=self.test.amqp_hostname,
                                 port=self.test.amqp_port,
                                 username=self.test.amqp_user,
                                 password=self.test.amqp_pwd,
                                 vhost=self.test.amqp_vhost,
                                 server_id='testqamserver',
                                 client_id=self.client_id,
                                 serializer='json'
                                 )

                for i in range(10):

                    str = self.client_id + '_%d' % i
                    golden_device = 'Hello ' + str
                    result = proxy.hello(str)
                    #logger.debug('GD: ' + golden_device + ';  Result: ' + result)
                    if result != golden_device:
                        #logger.debug(golden_device + ' !=  ' + result)
                        self.failed = True
                proxy.close()

        thread1 = TestClientThread('thread1', self)
        thread2 = TestClientThread('thread2', self)
        thread3 = TestClientThread('thread3', self)

        thread1.start()
        thread2.start()
        thread3.start()

        thread1.join()
        thread2.join()
        thread3.join()

        self.assertFalse(thread1.failed, "Thread1 failed")
        self.assertFalse(thread2.failed, "Thread1 failed")
    




    def test_multiple_method_threads(self):
        print 'test_multiple_method_threads'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testthreadedclient',
                         serializer='json'
                         )

        class TestClientThread(Thread):
            def __init__(self, client_id, proxy):
                Thread.__init__(self)
                self.proxy = proxy
                self.failed = False
                self.client_id = client_id
      
            def run(self):
                for i in range(10):

                    str = self.client_id + '_%d' % i
                    golden_device = 'Hello ' + str
                    result = self.proxy.hello(str)
                    #logger.debug('GD: ' + golden_device + ';  Result: ' + result)
                    if result != golden_device:
                        logger.debug(golden_device + ' !=  ' + result)
                        self.failed = True

        thread1 = TestClientThread('thread1', proxy)
        thread2 = TestClientThread('thread2', proxy)
        thread3 = TestClientThread('thread3', proxy)

        thread1.start()
        thread2.start()
        thread3.start()

        thread1.join()
        thread2.join()
        thread3.join()

        self.assertFalse(thread1.failed, "Thread1 failed")
        self.assertFalse(thread2.failed, "Thread2 failed")
        self.assertFalse(thread3.failed, "Thread3 failed")
        proxy.close()


  

    def test_exception_method_call(self):
        print 'test_exception_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception ',
                         serializer='json'
                         )

        golden_device = 'Help'
        gd2 = 24
        try:
            proxy.excp_method(golden_device, gd2)
        except Exception as e:
            arg = e.args[0]
            print repr(e)
            self.assertEqual(arg, golden_device, 'Exception Argument1 is not equal')
            arg = e.args[1]
            self.assertEqual(arg, gd2, 'Exception Argument2 is not equal')

        else:
            self.fail('Exception need to be raised to pass the test')
        proxy.close()

    def test_callback_method_call(self):
        print 'test_callback_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception',
                         serializer='json'
                         )

        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN

            def success(self, arg):
                #time.sleep(1)
                self.state = CallbackTest.STATE_SUCCESS
                print arg
            def error(self, arg):
                self.state = CallbackTest.STATE_ERROR
                print 'error:' + repr(arg)


        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error).callback_test_method("test")
        print 'UID: ' + uid

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state

        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state, "Callback must be successfull, but it isn't")
            #time.sleep(0.1)
        proxy.close()

    def test_register_class_method_call(self):
        print 'test_register_class_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception ',
                         serializer='json'
                         )
        a = 1
        b = 2
        gd = a + b
        result = proxy.testclass.test_method(a, b)
        self.assertEqual(result, gd, 'sums must be the same')
        proxy.close()

    def test_method_not_found(self):
        print 'test_method_not_found'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception ',
                         serializer='json'
                         )
        a = 1
        b = 2
        gd = a + b
        try:
            result = proxy.testclass.asdf(a, b)
            #self.assertEqual(result,gd,'sums must be the same')

        except QAMMethodNotFoundException, e:
            print e
            pass
        else:
            self.fail("method should not be found")

        finally:
            proxy.close()

    def test_callback_exception_method_not_found(self):
        print 'test_callback_exception_method_not_found'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception',
                         serializer='json'
                         )



        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN

            def success(self, arg):
                print arg
                print "ERROR"

            def error(self, arg):
                if isinstance(arg, QAMException):
                    self.state = CallbackTest.STATE_SUCCESS
                    print 'QAMException:' + repr(arg)
                else:
                    self.state = CallbackTest.STATE_ERROR
                    print 'No QAMException: ' + repr(arg)
                

        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error).asdf("test")

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state
        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state,
                         "Must return a QAM Exception because the method is not defined on the server")
        proxy.close()

    def test_callback_timedout(self):
        print 'test_callback_timedout'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception',
                         serializer='json'
                         )



        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN

            def success(self, arg):
                print arg
                self.state = CallbackTest.STATE_ERROR
                print "ERROR"

            def error(self, arg):
                if isinstance(arg, QAMTimeoutException):
                    self.state = CallbackTest.STATE_SUCCESS
                    print 'QAMException:' + repr(arg)
                else:
                    self.state = CallbackTest.STATE_ERROR
                    print 'No QAMException: ' + repr(arg)


        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error) \
                .set_timeout(1) \
                .looong_lasting_method()

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state
        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state,
                         "Must return a QAMTimeoutException because the \
                          method needs too much time to return result")
        proxy.close()

    def test_callback_exception_method_call(self):
        print 'test_callback_exception_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception',
                         serializer='json'
                         )

        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN

            def success(self, arg):
                print arg
                print "ERROR"

            def error(self, arg):
                if isinstance(arg, QAMException):
                    self.state = CallbackTest.STATE_ERROR
                    print 'QAMException:' + repr(arg)
                else:
                    self.state = CallbackTest.STATE_SUCCESS
                    print 'No QAMException: ' + repr(arg)
                    

        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error).callback_exception("test")

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state

        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state,
                 "Must return a User defined Exception to succeed.")
        proxy.close()
    
    def test_callback_instance_call(self):
        print 'test_callback_instance_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception',
                         serializer='json'
                         )

        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN
                self.result = 0

            def success(self, arg):
                print arg
                self.result = arg
                self.state = CallbackTest.STATE_SUCCESS

            def error(self, arg):
                if isinstance(arg, QAMException):
                    self.state = CallbackTest.STATE_ERROR
                    print 'QAMException:' + repr(arg)
                else:
                    self.state = CallbackTest.STATE_ERROR
                    print 'No QAMException: ' + repr(arg)


        a = 1
        b = 2
        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error) \
                .testclass.test_method(a, b)

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state

        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state,
                 "Must call the success method NOT the error method.")
        self.assertEqual(a+b, callback_test.result,
                 "Result does not match.")
        proxy.close()

    def test_callback_exception_method_not_found_in_instance(self):
        print 'test_callback_exception_method_not_found_in_instance'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testexception',
                         serializer='json'
                         )



        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN

            def success(self, arg):
                print arg
                print "ERROR"

            def error(self, arg):
                if isinstance(arg, QAMException):
                    self.state = CallbackTest.STATE_SUCCESS
                    print 'QAMException:' + repr(arg)
                else:
                    self.state = CallbackTest.STATE_ERROR
                    print 'No QAMException: ' + repr(arg)


        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error).testclass.asdf("test")

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state
        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state,
                         "Must return a QAM Exception because the method is not defined on the server")
        proxy.close()

    def test_synchronous_timedout_method_call(self):
        print 'test_synchronous_timedout_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserver',
                         client_id='testqamclient1',
                         serializer='json'
                         )

        success = False
        try:
            result = proxy.set_timeout(1).looong_lasting_method()
        except QAMTimeoutException:
            success = True


        self.assertTrue(success, "QAMTimoutException must be raised.")

        proxy.close()

    def test_serialize_pickle_class(self):
        print 'test_serialize_pickle'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserverpickle',
                         client_id='testqamclient1',
                         serializer='pickle'
                         )

        test_complex = ComplexTest()
        result = proxy.complex_type_method(test_complex)



        self.assertEqual(result, test_complex.var, "The result must be the same as TestComplex.var.")

        proxy.close()

    def test_serialize_pickle_exception(self):
        print 'test_serialize_pickle'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserverpickle',
                         client_id='testqamclient1',
                         serializer='pickle'
                         )

        exception = Exception("hallo")
        result = proxy.complex_method_exception(exception)



        self.assertEqual(result, exception[0], "The result must be the same as TestComplex.var.")

        proxy.close()

    def test_serialize_pickle_custom_exception(self):
        print 'test_serialize_pickle_custom_exception'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserverpickle',
                         client_id='testqamclient1',
                         serializer='pickle'
                         )

     
        success = False
        try:
            print "before call"
            result = proxy.complex_method_custom_exception()
        except Exception, e:
            if isinstance(e, CustomException):
                success = True



        self.assertTrue(success, "A CustomException must be thrown.")

        proxy.close()
        
    def test_callback_exception_method_call_pickle(self):
        print 'test_callback_exception_method_call_pickle'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserverpickle',
                         client_id='testexception',
                         serializer='pickle'
                         )

        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN

            def success(self, arg):
                print arg
                print "ERROR"
                self.state = CallbackTest.STATE_ERROR

            def error(self, arg):
                if isinstance(arg, CustomException):
                    self.state = CallbackTest.STATE_SUCCESS
                    print 'CustomException:' + repr(arg)
                else:
                    self.state = CallbackTest.STATE_ERROR
                    print 'No CustomException: ' + repr(arg)


        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error).complex_method_custom_exception()

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state

        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state,
                 "Must return a User defined Exception to succeed.")
        proxy.close()

    def test_asynchronous_timedout_method_call_pickle(self):
        print 'test_synchronous_timedout_method_call'
        proxy = QAMProxy(hostname=self.amqp_hostname,
                         port=self.amqp_port,
                         username=self.amqp_user,
                         password=self.amqp_pwd,
                         vhost=self.amqp_vhost,
                         server_id='testqamserverpickle',
                         client_id='testqamclient1',
                         serializer='pickle'
                         )

        class CallbackTest():
            (STATE_UNKNOWN, STATE_SUCCESS, STATE_ERROR) = range(3)
            def __init__(self):
                self.state = CallbackTest.STATE_UNKNOWN

            def success(self, arg):
                print arg
                print "ERROR"
                self.state = CallbackTest.STATE_ERROR

            def error(self, arg):
                if isinstance(arg, QAMTimeoutException):
                    self.state = CallbackTest.STATE_SUCCESS
                    print 'CustomException:' + repr(arg)
                else:
                    self.state = CallbackTest.STATE_ERROR
                    print 'No CustomException: ' + repr(arg)


        callback_test = CallbackTest()
        uid = proxy.callback(callback_test.success, callback_test.error).set_timeout(1).delayed_exception_method()

        old_state = -1
        while True:
            state  = proxy.get_callback_state(uid)
            if state == 0 and old_state != 0: print "state: Waiting"
            if state == 1 and old_state != 1: print "state: Processing"
            if state == 2:
                print "state: Finished"
                break
            old_state = state

        self.assertEqual(CallbackTest.STATE_SUCCESS, callback_test.state,
                 "Must return a User defined Exception to succeed.")
        proxy.close()


if __name__ == '__main__':
    unittest.main()
#    result = unittest.TestResult()
#    test = QamTestCase('test_callback_exception_method_call_pickle')
#    test.run(result)
#    if len(result.failures) != 0: print "=============== Failures ==========================="
#    print "".join("%s = %s" % (a, b) for a, b in result.failures)
#    if len(result.errors) != 0: print "=============== Errors ============================="
#    print "".join("%s = %s" % (a, b) for a, b in result.errors)
#    print "===================================================="
#    if result.wasSuccessful(): print 'Successfully passed all Tests'
#    else: print "Failed - Details see above."



from setuptools import setup, find_packages
import codecs
import sys, os
sys.path.insert(0,'src/')
init_pyc = 'src/qam/__init__.pyc'
if os.path.exists(init_pyc):
    os.remove(init_pyc)

import qam


if os.path.exists("doc/source/introduction.rst"):
    long_description = codecs.open('doc/source/introduction.rst', "r", "utf-8").read()
else:
    long_description = "See " + qam.__homepage__

setup(
    name = "qam",
    version = qam.__version__,
    packages = find_packages('src'),
    package_dir = {'':'src'},
    install_requires = ['carrot>=0.5'],
  
    
    # metadata for upload to PyPI
    author = qam.__author__,
    author_email = qam.__contact__,
    description = qam.__doc__,
    long_description=long_description,
    keywords = "rpc amqp",
    platforms=["any"],
    url = qam.__homepage__,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Topic :: Communications",
        "Topic :: System :: Distributed Computing",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
 

)

